from os.path import join as pjoin
from appdirs import AppDirs
import os
import gc
from nltk.corpus.util import _make_bound_method


class ResourceMan:
    RESOURCE_NAME = None

    def __init__(self):
        self._dirs = AppDirs("finntk", "Frankie Robertson")
        self._resources = {}

    def _get_data_dir(self):
        if "FINNTK_DATA" in os.environ:
            dir = os.environ["FINNTK_DATA"]
        else:
            dir = self._dirs.user_data_dir
        if self.RESOURCE_NAME:
            dir = pjoin(dir, self.RESOURCE_NAME)
        return dir

    def _ensure_data_dir(self):
        dir = self._get_data_dir()
        if not os.path.exists(dir):
            os.makedirs(dir)
        return dir

    def _get_res_filename(self, res):
        return self._resources[res]

    def _get_res_path(self, res):
        data_dir = self._ensure_data_dir()
        filename = self._get_res_filename(res)
        return pjoin(data_dir, filename)

    def get_res(self, res):
        path = self._get_res_path(res)
        if not os.path.exists(path):
            self._bootstrap(path)
        return path

    def bootstrap(self):
        if not self.RESOURCE_NAME:
            assert False
        dir = self._get_data_dir()
        if not os.path.exists(dir):
            self._bootstrap()


# Taken from nltk with data file management stuff taken out
class LazyCorpusLoader:

    def __init__(self, name, reader_cls, *args, **kwargs):
        from nltk.corpus.reader.api import CorpusReader

        assert issubclass(reader_cls, CorpusReader)
        self.__name = self.__name__ = name
        self.__reader_cls = reader_cls
        self.__args = args
        self.__kwargs = kwargs

    def __load(self):
        # Load the corpus.
        corpus = self.__reader_cls("", *self.__args, **self.__kwargs)

        # This is where the magic happens!  Transform ourselves into
        # the corpus by modifying our own __dict__ and __class__ to
        # match that of the corpus.

        args, kwargs = self.__args, self.__kwargs
        name, reader_cls = self.__name, self.__reader_cls

        self.__dict__ = corpus.__dict__
        self.__class__ = corpus.__class__

        # _unload support: assign __dict__ and __class__ back, then do GC.
        # after reassigning __dict__ there shouldn't be any references to
        # corpus data so the memory should be deallocated after gc.collect()
        def _unload(self):
            lazy_reader = LazyCorpusLoader(name, reader_cls, *args, **kwargs)
            self.__dict__ = lazy_reader.__dict__
            self.__class__ = lazy_reader.__class__
            gc.collect()

        self._unload = _make_bound_method(_unload, self)

    def __getattr__(self, attr):

        # Fix for inspect.isclass under Python 2.6
        # (see http://bugs.python.org/issue1225107).
        # Without this fix tests may take extra 1.5GB RAM
        # because all corpora gets loaded during test collection.
        if attr == "__bases__":
            raise AttributeError("LazyCorpusLoader object has no attribute '__bases__'")

        self.__load()
        # This looks circular, but its not, since __load() changes our
        # __class__ to something new:
        return getattr(self, attr)

    def __repr__(self):
        return "<%s in %r (not loaded yet)>" % (
            self.__reader_cls.__name__, ".../corpora/" + self.__name
        )

    def _unload(self):
        # If an exception occures during corpus loading then
        # '_unload' method may be unattached, so __getattr__ can be called;
        # we shouldn't trigger corpus loading again in this case.
        pass



import subprocess as sp
import multiprocessing as mp
from finntk.omor.anlys import pairs_to_dict


def analysis_to_pairs(ana):
    for bit in ana.split("]|["):
        k, v = bit.strip("[]").split("=", 1)
        yield k, v


def parse_finnpos_line(line):
    surf, _, lemma, feats, _ = line.split("\t")
    return surf, lemma, feats


def analysis_to_dict(ana):
    return pairs_to_dict(analysis_to_pairs(ana))


def batch_finnpos(source_iter, *args, maxsize=0, **kwargs):
    """
    Shovel sentences through FinnPOS. A typical use case for this would be when
    you want to tag a bunch of sentences with FinnPOS at the same time as doing
    some other kind of transformation.

    `source_iter` should be an iterator returning pairs `(sent, extra)` where
    sent is a list of tokens and `extra` is any sentence identifier you would
    like to pass through. It will be run in a new process. If you would like to
    pass arguments to it, pass them in as extra arguments to `batch_finnpos`.
    """
    done_sentinel = object()

    def source_func(finnpos, queue, *args, **kwargs):
        for sent, extra in source_iter(*args, **kwargs):
            finnpos.feed_sent(sent)
            queue.put(extra)
        queue.put(done_sentinel)

    with FinnPOSCtx() as finnpos:
        ctx = mp.get_context("fork")
        id_queue = ctx.Queue(maxsize=0)
        source_proc = ctx.Process(
            target=source_func, args=(finnpos, id_queue) + args, kwargs=kwargs
        )
        source_proc.start()
        while 1:
            extra = id_queue.get()
            if extra is done_sentinel:
                break
            tagged_sent = finnpos.get_analys()
            yield tagged_sent, extra
        source_proc.join()


class FinnPOS():

    def __init__(self):
        self.proc = sp.Popen(
            ["ftb-label"], stdin=sp.PIPE, stdout=sp.PIPE, universal_newlines=True
        )

    def cleanup(self):
        if not self.proc.stdin.closed:
            self.proc.stdin.close()

    def feed_sent(self, sent):
        for token in sent:
            self.proc.stdin.write(token)
            self.proc.stdin.write("\n")
        self.proc.stdin.write("\n")
        self.proc.stdin.flush()

    def get_analys(self):
        tagged_sent = []
        for line in self.proc.stdout:
            if line == "\n":
                break
            surf, lemma, feats_str = parse_finnpos_line(line[:-1])
            feats = analysis_to_dict(feats_str)
            tagged_sent.append((surf, lemma, feats))
        return tagged_sent

    def __del__(self):
        self.cleanup()

    def __call__(self, sent):
        """
        Transform a single sentence with FinnPOS.

        Note that using this repeatedly serialises your processing pipeline
        sentence-by-sentence. If performance is a concern, consider using
        `batch_finnpos` if possible in this situation.
        """
        self.feed_sent(sent)
        return self.get_analys()


class FinnPOSCtx():
    """
    This helper lets you get an instance of `FinnPOS` and ensures it is
    correctly cleaned up. Usually you should use this instead of instantiating
    FinnPOS directly.
    """

    def __enter__(self):
        self.finnpos = FinnPOS()

    def __exit__(self):
        self.finnpos.cleanup()
        del self.finnpos


_global_finnpos = None


def sent_finnpos(sent):
    FinnPOS.__call__.__doc__ + """

    This function will keep a single global copy of FinnPOS running.

    Note that this function is not thread safe and is a convenience for
    exploratory programming only. The recommended method is to use FinnPOSCtx.
    """
    global _global_finnpos
    if _global_finnpos is None:
        _global_finnpos = FinnPOS()
    return _global_finnpos(sent)


def cleanup():
    """
    Cleanup the global FinnPOS instance kept by `sent_finnpos`. If you want to
    use this, you should consider using `FinnPOSCtx` instead.
    """
    global _global_finnpos
    _global_finnpos = None



from scipy.spatial.distance import cosine
from nltk.tokenize import word_tokenize
from gensim.models import KeyedVectors
import logging
from urllib.request import urlretrieve
import os
from finntk.utils import ResourceMan

logger = logging.getLogger(__name__)


class WordVecs(ResourceMan):
    RESOURCE_NAME = "fasttext-multilingual"

    FI_URL = "https://s3.amazonaws.com/arrival/embeddings/wiki.multi.fi.vec"
    EN_URL = "https://s3.amazonaws.com/arrival/embeddings/wiki.multi.en.vec"

    def __init__(self):
        super().__init__()
        self._resources["fi_vec"] = "wiki.multi.fi.binvec"
        self._resources["en_vec"] = "wiki.multi.en.binvec"
        self._en = None
        self._fi = None

    def _download(self, lang, url, dest):
        logger.info("Downloading {} word vectors".format(lang))
        tmp_fn, _ = urlretrieve(url)
        try:
            logger.info("Converting {} word vectors".format(lang))
            fi = KeyedVectors.load_word2vec_format(tmp_fn)
            fi.save(dest)
        finally:
            os.remove(tmp_fn)

    def _bootstrap(self, _res):
        self._download("Finnish", self.FI_URL, self._get_res_path("en_vec"))
        self._download("English", self.EN_URL, self._get_res_path("fi_vec"))

    def get_en(self):
        if self._en is None:
            en_vec_path = self.get_res("en_vec")
            logger.info("Loading English word vectors")
            self._en = KeyedVectors.load(en_vec_path, mmap="r")
            logger.info("Loaded English word vectors")
        return self._en

    def get_fi(self):
        if self._fi is None:
            fi_vec_path = self.get_res("fi_vec")
            logger.info("Loading Finnish word vectors")
            self._fi = KeyedVectors.load(fi_vec_path, mmap="r")
            logger.info("Loaded Finnish word vectors")
        return self._fi


word_vecs = WordVecs()


def avg_vec(space, stream):
    word_count = 0
    vec_sum = 0
    for word in stream:
        try:
            vec = space.get_vector(word)
        except KeyError:
            continue
        else:
            vec_sum += vec
            word_count += 1
    if word_count == 0:
        return None
    return vec_sum / word_count


def mk_context_vec(word_forms):
    fi = word_vecs.get_fi()
    return avg_vec(fi, word_forms)


def get_defn_distance(context_vec, defn):
    en = word_vecs.get_en()
    defn_vec = avg_vec(en, word_tokenize(defn))
    if defn_vec is None or context_vec is None:
        return 7
    return cosine(defn_vec, context_vec)


def disambg(lemmas, context):
    context_vec = mk_context_vec(context)
    best_lemma = None
    best_dist = 8
    for lemma in lemmas:
        dist = get_defn_distance(context_vec, lemma.synset().definition())
        if dist < best_dist:
            best_lemma = lemma
            best_dist = dist
    return best_lemma, best_dist



from finntk.utils import ResourceMan, LazyCorpusLoader
from os.path import join as pjoin
from nltk.corpus import wordnet
from nltk.corpus.reader.wordnet import (
    WordNetCorpusReader,
    Synset,
    Lemma,
    VERB_FRAME_STRINGS,
    ADJ,
    ADJ_SAT,
    WordNetError,
)
import logging
import re
from plumbum.cmd import git

logger = logging.getLogger(__name__)


# This lemma is broken in data.noun
LEMMA_NAME_FIXES = {"moderniusAdditional_synonym": "modernius"}

# These lemmas occur in differently upper/lower case wise in transls and data.POS
IGNORE_CASE_LEMMAS = {"ci", "otc", "rh-negatiivinen", "rh-positiivinen"}


class FinnWordNetResMan(ResourceMan):
    RESOURCE_NAME = "fiwn"
    RELS = ["transls"]

    REPO = "https://github.com/frankier/fiwn.git"

    def __init__(self):
        super().__init__()
        for name in self.RELS:
            self._resources[name] = pjoin("data", "rels", "fiwn-{}.tsv".format(name))
        self._resources["synset_map"] = "synset_map.tsv"

    def _get_res_filename(self, res):
        return self._resources.get(res, pjoin("data", "dict", res))

    def _bootstrap(self, _res=None):
        data_dir = self._get_data_dir()
        git("clone", self.REPO, data_dir)


fiwn_resman = FinnWordNetResMan()


class FinnWordNetReader(WordNetCorpusReader):

    def __init__(self, *args, **kwargs):
        fiwn_resman.bootstrap()
        args = list(args)
        args[0] = fiwn_resman.get_res("")
        super().__init__(*args, **kwargs)

    def open(self, filename):
        if filename == "lexnames":
            return wordnet.open("lexnames")
        return super().open(filename)

    def _synset_from_pos_and_line(self, pos, data_file_line):
        # Copied and modified for FiWN. The only change is the handling of
        # FiWN's special <tag/> syntactic markers.

        # Construct a new (empty) synset.
        synset = Synset(self)

        # parse the entry for this synset
        try:

            # parse out the definitions and examples from the gloss
            columns_str, gloss = data_file_line.split("|")
            gloss = gloss.strip()
            definitions = []
            for gloss_part in gloss.split(";"):
                gloss_part = gloss_part.strip()
                if gloss_part.startswith('"'):
                    synset._examples.append(gloss_part.strip('"'))
                else:
                    definitions.append(gloss_part)
            synset._definition = "; ".join(definitions)

            # split the other info into fields
            _iter = iter(columns_str.split())

            def _next_token():
                return next(_iter)

            # get the offset
            synset._offset = int(_next_token())

            # determine the lexicographer file name
            lexname_index = int(_next_token())
            synset._lexname = self._lexnames[lexname_index]

            # get the part of speech
            synset._pos = _next_token()

            # create Lemma objects for each lemma
            n_lemmas = int(_next_token(), 16)
            for _ in range(n_lemmas):
                # get the lemma name
                lemma_name = _next_token()
                # get the lex_id (used for sense_keys)
                lex_id = int(_next_token(), 16)
                # If the lemma has a syntactic marker, extract it.
                m = re.match(r"([^<]*)(<.*)?$", lemma_name)
                lemma_name, syn_mark = m.groups()
                # create the lemma object
                lemma = Lemma(self, synset, lemma_name, lexname_index, lex_id, syn_mark)
                synset._lemmas.append(lemma)
                synset._lemma_names.append(lemma._name)

            # collect the pointer tuples
            n_pointers = int(_next_token())
            for _ in range(n_pointers):
                symbol = _next_token()
                offset = int(_next_token())
                pos = _next_token()
                lemma_ids_str = _next_token()
                if lemma_ids_str == "0000":
                    synset._pointers[symbol].add((pos, offset))
                else:
                    source_index = int(lemma_ids_str[:2], 16) - 1
                    target_index = int(lemma_ids_str[2:], 16) - 1
                    source_lemma_name = synset._lemmas[source_index]._name
                    lemma_pointers = synset._lemma_pointers
                    tups = lemma_pointers[source_lemma_name, symbol]
                    tups.append((pos, offset, target_index))

            # read the verb frames
            try:
                frame_count = int(_next_token())
            except StopIteration:
                pass
            else:
                for _ in range(frame_count):
                    # read the plus sign
                    plus = _next_token()
                    assert plus == "+"
                    # read the frame and lemma number
                    frame_number = int(_next_token())
                    frame_string_fmt = VERB_FRAME_STRINGS[frame_number]
                    lemma_number = int(_next_token(), 16)
                    # lemma number of 00 means all words in the synset
                    if lemma_number == 0:
                        synset._frame_ids.append(frame_number)
                        for lemma in synset._lemmas:
                            lemma._frame_ids.append(frame_number)
                            lemma._frame_strings.append(frame_string_fmt % lemma._name)
                    # only a specific word in the synset
                    else:
                        lemma = synset._lemmas[lemma_number - 1]
                        lemma._frame_ids.append(frame_number)
                        lemma._frame_strings.append(frame_string_fmt % lemma._name)

        # raise a more informative error with line text
        except ValueError as e:
            raise WordNetError("line %r: %s" % (data_file_line, e))

        # set sense keys for Lemma objects - note that this has to be
        # done afterwards so that the relations are available
        for lemma in synset._lemmas:
            if synset._pos == ADJ_SAT:
                head_lemma = synset.similar_tos()[0]._lemmas[0]
                head_name = head_lemma._name
                head_id = "%02d" % head_lemma._lex_id
            else:
                head_name = head_id = ""
            tup = (
                lemma._name,
                WordNetCorpusReader._pos_numbers[synset._pos],
                lemma._lexname_index,
                lemma._lex_id,
                head_name,
                head_id,
            )
            lemma._key = ("%s%%%d:%02d:%02d:%s:%s" % tup).lower()

        # the canonical name is based on the first lemma
        lemma_name = synset._lemmas[0]._name.lower()
        offsets = self._lemma_pos_offset_map[lemma_name][synset._pos]
        sense_index = offsets.index(synset._offset)
        tup = lemma_name, synset._pos, sense_index + 1
        synset._name = "%s.%s.%02i" % tup

        return synset

    def _load_lemma_pos_offset_map(self):
        # Copied and modified for FiWN. The only changes are to handle a single
        # entry: FiWN contains an empty lemma for `taken` as in `taken
        # ill/drunk`. Thus lines beginning with two spaces as are treated as a
        # comment instead of ones starting with one space and tokenisation is
        # done on spaces instead of all whitespace.

        for suffix in self._FILEMAP.values():

            # parse each line of the file (ignoring comment lines)
            for i, line in enumerate(self.open("index.%s" % suffix)):
                if line.startswith("  "):
                    continue

                _iter = iter(line.split(" "))

                def _next_token():
                    return next(_iter)

                try:

                    # get the lemma and part-of-speech
                    lemma = _next_token()
                    pos = _next_token()

                    # get the number of synsets for this lemma
                    n_synsets = int(_next_token())
                    assert n_synsets > 0

                    # get and ignore the pointer symbols for all synsets of
                    # this lemma
                    n_pointers = int(_next_token())
                    [_next_token() for _ in range(n_pointers)]

                    # same as number of synsets
                    n_senses = int(_next_token())
                    assert n_synsets == n_senses

                    # get and ignore number of senses ranked according to
                    # frequency
                    _next_token()

                    # get synset offsets
                    synset_offsets = [int(_next_token()) for _ in range(n_synsets)]

                # raise more informative error with file name and line number
                except (AssertionError, ValueError) as e:
                    tup = ("index.%s" % suffix), (i + 1), e
                    raise WordNetError("file %s, line %i: %s" % tup)

                # map lemmas and parts of speech to synsets
                self._lemma_pos_offset_map[lemma][pos] = synset_offsets
                if pos == ADJ:
                    self._lemma_pos_offset_map[lemma][ADJ_SAT] = synset_offsets


fiwn = LazyCorpusLoader("fiwn", FinnWordNetReader, None)


class EnCountsFinnWordNetReader(FinnWordNetReader):

    def lemma_count(self, lemma):
        if not hasattr(self, "_counts"):
            self._counts = calc_fiwn_counts()
        return self._counts.get(lemma.key(), 0.0)

    def lemmas(self, lemma, pos=None, lang="eng"):
        lemmas = super().lemmas(lemma, pos, lang)
        return sorted(lemmas, key=lambda l: -l.count())


fiwn_encnt = LazyCorpusLoader("fiwn", EnCountsFinnWordNetReader, None)


en_fi_maps = None


def get_en_fi_maps():

    def fix_synset_key(synset):
        synset = synset.split(":", 1)[1]
        if synset[0] == "s":
            synset = "a" + synset[1:]
        return synset

    global en_fi_maps
    if en_fi_maps is None:
        synset_map = fiwn_resman.get_res("synset_map")
        fi2en = {}
        en2fi = {}
        for line in open(synset_map):
            en_synset_key, fi_synset_key = line[:-1].split("\t")
            en_synset_key = fix_synset_key(en_synset_key)
            fi_synset_key = fix_synset_key(fi_synset_key)
            fi2en[fi_synset_key] = en_synset_key
            en2fi[en_synset_key] = fi_synset_key
        en_fi_maps = fi2en, en2fi
    return en_fi_maps


def get_transl_iter():
    transls = fiwn_resman.get_res("transls")
    for line in open(transls):
        fi_synset, fi_lemma, en_synset, en_lemma, rel, extra = line[:-1].split("\t")
        _, fi_synset = fi_synset.split(":", 1)
        _, en_synset = en_synset.split(":", 1)
        fi_lemma = fi_lemma.split("<", 1)[0]
        yield fi_synset, fi_lemma, en_synset, en_lemma, rel, extra


def surf_to_norm_lemma(surf_lemma):
    return surf_lemma.replace(" ", "_").replace(",", "\\,").replace("(", "\\(").replace(
        ")", "\\)"
    )


def synset_surf_to_lemma(synset, surf):
    lemmas = synset.lemmas()
    for lemma in lemmas:
        normed_lemma = surf_to_norm_lemma(surf)
        if (
            normed_lemma.lower() in IGNORE_CASE_LEMMAS
            and normed_lemma.lower() == lemma.name().lower()
        ):
            return lemma
        fixed_lemma_name = LEMMA_NAME_FIXES.get(lemma.name(), lemma.name())
        if fixed_lemma_name == normed_lemma:
            return lemma


def get_lemma(wn, synset_key, lemma_str):
    pos = synset_key[0]
    offset = int(synset_key[1:], 10)
    synset = wn.synset_from_pos_and_offset(pos, offset)
    return synset_surf_to_lemma(synset, lemma_str)


def calc_fiwn_counts():
    en2fi = {}
    for (
        fi_synset_key, fi_lemma_str, en_synset_key, en_lemma_str, rel, extra
    ) in get_transl_iter():
        if rel != "synonym":
            continue

        fi_lemma = get_lemma(fiwn, fi_synset_key, fi_lemma_str)
        assert fi_lemma is not None

        en_lemma = get_lemma(wordnet, en_synset_key, en_lemma_str)
        assert en_lemma is not None

        en2fi.setdefault(en_lemma.key(), []).append(fi_lemma.key())
    counts = {}
    for en, fis in en2fi.items():
        for fi in fis:
            counts.setdefault(fi, 0.0)
            try:
                en_lemma = wordnet.lemma_from_key(en)
            except WordNetError:
                # The following lemmas are not in the PWN sense index for some reason:
                # ['earth%1:17:02::', 'ddc%1:06:01::', 'kb%1:23:01::', 'sun%1:17:02::',
                # 'moon%1:17:03::', 'earth%1:15:01::', 'ddi%1:06:01::', 'kb%1:23:03::']
                pass
            else:
                counts[fi] += en_lemma.count() / len(fis)
    return counts



import ahocorasick

_abbrv_auto = None

PLACEHOLDER_ABBRVS = [
    "jstak",  # Abbreviation of jostakin.
    "jssak",  # Abbreviation of jossakin.
    "jnak",  # Abbreviation of jonakin.
    "jtak",  # Abbreviation of jotakin.
    "jllak",  # Abbreviation of jollakin.
    "jltak",  # Abbreviation of joltakin.
    "jksta",  # Abbreviation of jostakusta.
    "jllek",  # Abbreviation of jollekin.
    "jkssa",  # Abbreviation of jossakussa.
    "jkta",  # Abbreviation of jotakuta.
    "jklla",  # Abbreviation of jollakulla.
    "jklta",  # Abbreviation of joltakulta.
    "jhk",  # Abbreviation of johonkin.
    "jklle",  # Abbreviation of jollekulle.
    "jksik",  # Abbreviation of joksikin.
    "jkksi",  # Abbreviation of joksikuksi.
    "jkna",  # Abbreviation of jonakuna.
    "jkhun",  # Abbreviation of johonkuhun.
    "jku",  # Abbreviation of joku (“somebody”).
    "jk",  # Abbreviation of jokin (“something”).
    "jkn",  # Abbreviation of jonkun (“of somebody, somebody's”).
    "jnk",  # Abbreviation of jonkin.
]


def get_abbrv_auto():
    global _abbrv_auto
    if _abbrv_auto is not None:
        return _abbrv_auto
    _abbrv_auto = ahocorasick.Automaton()
    for abbrv in PLACEHOLDER_ABBRVS:
        _abbrv_auto.add_word("_{}_".format(abbrv), abbrv)
    _abbrv_auto.make_automaton()
    return _abbrv_auto


def has_abbrv(lemma):
    """
    Given a FinnWordNet formatted lemma, e.g. saada_tehdä_jtak return whether
    it contains a placeholder abbreviation.
    """
    abbrv_auto = get_abbrv_auto()
    it = abbrv_auto.iter("_{}_".format(lemma))
    return len(list(it))
